import os
from transformers import Seq2SeqTrainingArguments, Seq2SeqTrainer, PushToHubCallback
from .collator import DataCollatorSpeechSeq2SeqWithPadding
import evaluate
from datasets import DatasetDict
from .wandb_callback import WandbProgressResultsCallback
from transformers.trainer_pt_utils import IterableDatasetShard
from torch.utils.data import IterableDataset
from transformers import TrainerCallback
from .whisper_model_prep import WhisperModelPrep

class ShuffleCallback(TrainerCallback):
    def on_epoch_begin(self, args, state, control, train_dataloader, **kwargs):
        if isinstance(train_dataloader.dataset, IterableDatasetShard):
            pass 
        elif isinstance(train_dataloader.dataset, IterableDataset):
            train_dataloader.dataset.set_epoch(train_dataloader.dataset._epoch + 1)

class Trainer:
    """
    
    A Trainer class for fine-tuning and training speech-to-text models using the Hugging Face Transformers library.

    """
    def __init__(self, huggingface_write_token:str, model_id: str, dataset: DatasetDict , model: str, feature_processor, feature_extractor, tokenizer, language_abbr: str,wandb_api_key: str, use_peft:bool):
        """
        Initializes the Trainer with the necessary components and configurations for training.

        Parameters:
            huggingface_push_token (str): Hugging Face API token for authenticated push access.
            model_id (str): Identifier for the pre-trained model.
            dataset (DatasetDict): The dataset split into 'train' and 'test'.
            model (PreTrainedModel): The model instance to be trained.
            feature_processor (Any): The audio feature processor.
            feature_extractor (Any): The audio feature extractor.
            tokenizer (PreTrainedTokenizer): The tokenizer for text data.
            language_abbr (str): Abbreviation for the dataset's language.
        """
        os.environ["WANDB_API_KEY"] = wandb_api_key
        self.dataset = dataset
        self.model = model
        self.model_id = model_id
        self.tokenizer = tokenizer
        self.feature_processor = feature_processor
        self.feature_extractor = feature_extractor
        self.huggingface_write_token = huggingface_write_token
        self.language_abbr = language_abbr
        self.use_peft = use_peft

    def compute_metrics(self, pred) -> dict:
        """
        Computes the Word Error Rate (WER) metric for the model predictions.

        Parameters:
            pred (PredictionOutput): The output from the model's prediction.

        Returns:
            dict: A dictionary containing the computed WER metric.
        """
        pred_ids = pred.predictions
        label_ids = pred.label_ids
        label_ids[label_ids == -100] = self.tokenizer.pad_token_id
        pred_str = self.tokenizer.batch_decode(pred_ids, skip_special_tokens=True, normalize = True)
        label_str = self.tokenizer.batch_decode(label_ids, skip_special_tokens=True, normalize = True)
        metric = evaluate.load("wer")
        wer = 100 * metric.compute(predictions=pred_str, references=label_str)
        return {"wer": wer}
    
    def compute_spectrograms(self, example) ->  dict:
        waveform =  example["audio"]["array"]
        model_prep = WhisperModelPrep(self.dataset, self.model_id, self.language_abbr, 'transcribe', self.use_peft)
        feature_extractor = model_prep.initialize_feature_extractor()
        specs = feature_extractor(waveform, sampling_rate=16000, padding="do_not_pad").input_features[0]
        return {"spectrogram": specs}

    def train(self):
        """
        
        Conducts the training process using the specified model, dataset, and training configurations.
        
        """
        data_collator = DataCollatorSpeechSeq2SeqWithPadding(processor=self.feature_processor)
        training_args = Seq2SeqTrainingArguments(
            output_dir=f"./{self.model_id}-{self.language_abbr}",  
            per_device_train_batch_size=64,
            gradient_accumulation_steps=1,
            learning_rate=1e-5,
            warmup_steps=100,
            max_steps=200,
            gradient_checkpointing=True,
            fp16=False,
            optim="adamw_bnb_8bit",
            evaluation_strategy="steps",
            per_device_eval_batch_size=32,
            predict_with_generate=True,
            generation_max_length=225,
            save_steps=50,
            eval_steps=50,
            logging_steps=100,
            load_best_model_at_end=True,
            metric_for_best_model="wer",
            greater_is_better=False,
            push_to_hub = True,
            hub_token = self.huggingface_write_token,
            report_to = "wandb",
            remove_unused_columns=False, 
            ignore_data_skip=True
        )

        eval_dataset = self.dataset["test"].map(self.compute_spectrograms)

        trainer = Seq2SeqTrainer(
            args=training_args,
            model=self.model,
            train_dataset=self.dataset["train"],
            eval_dataset=eval_dataset,
            data_collator=data_collator,
            compute_metrics=self.compute_metrics,
            tokenizer=self.feature_processor.feature_extractor,
            callbacks=[ShuffleCallback()],            
        )   

        model_prep = WhisperModelPrep(self.dataset, self.model_id, self.language_abbr, 'transcribe', self.use_peft)
        tokenizer = model_prep.initialize_tokenizer()
        tokenizer.save_pretrained(training_args.output_dir)

        # trainer.add_callback(PushToHubCallback(output_dir=training_args.output_dir, tokenizer=tokenizer, hub_token = training_args.hub_token))

        progress_callback = WandbProgressResultsCallback(trainer, eval_dataset, tokenizer)
        trainer.add_callback(progress_callback)
        trainer.train()


from .demo_inference import WhisperDemo
import argparse


def parse_args():
    parser = argparse.ArgumentParser(description="Whisper Demo: Transcribe Audio and YouTube")
    parser.add_argument("--model_name", type=str, help="Name of the fine-tuned model to use in your huggingfacehub repo")
    parser.add_argument("--language_abbr", type=str, help="Language abbreviation for transcription")
    parser.add_argument("--tokenizer", type = str, help = "Whisper model version you used to fine-tune your model e.g openai/whisper-tiny, openai/whisper-base, openai/whisper-small, openai/whisper-medium, openai/whisper-large, openai/whisper-large-v2")
    parser.add_argument("--huggingface_read_token", type = str, help = "Hugging Face API token for read authenticated access.")
    return parser.parse_args()



if __name__ == "__main__":
    args = parse_args()
    demo = WhisperDemo(args.model_name, args.language_abbr, args.tokenizer, args.huggingface_read_token)
    demo.generate_demo()



import evaluate

class MetricComputer:
    """
    A class to compute metrics for model predictions.

    This class encapsulates the functionality of loading a specific metric from the
    `evaluate` library and computing it using predictions and labels processed by a tokenizer.

    """

    def __init__(self, metric_name: str, tokenizer):
        """
        Initializes the MetricComputer with a specified metric and tokenizer.

        Args:
            metric_name (str): The name of the metric to load (e.g., "wer" for word error rate).
            tokenizer: The tokenizer to use for decoding prediction and label IDs.
        """
        self.metric = evaluate.load(metric_name)
        self.tokenizer = tokenizer

    def compute_metrics(self, pred) -> dict:
        """
        Computes the metric for a set of predictions and labels.

        The function processes the predictions and labels, replacing any padding token IDs with -100,
        and decodes them using the provided tokenizer before computing the metric.

        Args:
            pred: An object that contains `predictions` and `label_ids`, which are used to compute the metric.
                  The `predictions` and `label_ids` should be arrays of IDs.

        Returns:
            dict: A dictionary containing the computed metric value.
        """
        pred_ids = pred.predictions
        label_ids = pred.label_ids
        label_ids[label_ids == -100] = self.tokenizer.pad_token_id
        pred_str = self.tokenizer.batch_decode(pred_ids, skip_special_tokens=True)
        label_str = self.tokenizer.batch_decode(label_ids, skip_special_tokens=True)
        wer = 100 * self.metric.compute(predictions=pred_str, references=label_str)

        return {"wer": wer}





import torch
import gradio as gr
import pytube as pt
from transformers import pipeline, WhisperTokenizer

class WhisperDemo:
    def __init__(self, model_name, language_abbr, tokenizer, huggingface_read_token):
        self.model_name = model_name
        self.language_abbr = language_abbr
        self.tokenizer = tokenizer
        self.huggingface_read_token = huggingface_read_token
        self.pipe = None

    def initialize_pipeline(self):
        tokenizer = WhisperTokenizer.from_pretrained(self.tokenizer)
        device = 0 if torch.cuda.is_available() else "cpu"
        self.pipe = pipeline(
            task="automatic-speech-recognition",
            model=self.model_name,
            token=self.huggingface_read_token,
            tokenizer=tokenizer,
            chunk_length_s=30,
            device=device,
        )
        self.pipe.model.config.forced_decoder_ids = self.pipe.tokenizer.get_decoder_prompt_ids(language=self.language_abbr, task="transcribe")

    def transcribe(self, microphone, file_upload):
        warn_output = ""
        if (microphone is not None) and (file_upload is not None):
            warn_output = (
                "WARNING: You've uploaded an audio file and used the microphone. "
                "The recorded file from the microphone will be used and the uploaded audio will be discarded.\n"
            )

        elif (microphone is None) and (file_upload is None):
            return "ERROR: You have to either use the microphone or upload an audio file"

        file = microphone if microphone is not None else file_upload
        text = self.pipe(file)["text"]

        return warn_output + text

    def yt_transcribe(self, yt_url):
        yt = pt.YouTube(yt_url)
        video_id = yt_url.split("?v=")[-1]
        html_embed_str = (
            f'<center> <iframe width="500" height="320" src="https://www.youtube.com/embed/{video_id}"> </iframe>'
            " </center>"
        )
        stream = yt.streams.filter(only_audio=True)[0]
        stream.download(filename="audio.mp3")
        text = self.pipe("audio.mp3")["text"]

        return html_embed_str, text

    def generate_demo(self):
        self.initialize_pipeline()

        mf_transcribe = gr.Interface(
            fn=self.transcribe,
            inputs=[
                gr.Audio(sources="microphone", type="filepath"),
                gr.Audio(sources="upload", type="filepath"),
            ],
            outputs="text",
            title="Whisper Demo: Transcribe Audio",
            description=(
                "Transcribe long-form microphone or audio inputs with the click of a button! Demo uses the fine-tuned"
                f" checkpoint [{self.model_name}](https://huggingface.co/{self.model_name}) and 🤗 Transformers to transcribe audio files"
                " of arbitrary length."
            ),
            allow_flagging="never",
        )

        yt_transcribe_interface = gr.Interface(
            fn=self.yt_transcribe,
            inputs=[gr.Textbox(lines=1, placeholder="Paste the URL to a YouTube video here", label="YouTube URL")],
            outputs=["html", "text"],
            title="Whisper Demo: Transcribe YouTube",
            description=(
                "Transcribe long-form YouTube videos with the click of a button! Demo uses the fine-tuned checkpoint:"
                f" [{self.model_name}](https://huggingface.co/{self.model_name}) and 🤗 Transformers to transcribe audio files of"
                " arbitrary length."
            ),
            allow_flagging="never",
        )

        demo = gr.TabbedInterface([mf_transcribe, yt_transcribe_interface], ["Transcribe Audio", "Transcribe YouTube"])
        demo.launch(share=True)



from transformers import WhisperFeatureExtractor, WhisperTokenizer, WhisperProcessor
from datasets import DatasetDict
from transformers import WhisperForConditionalGeneration
from peft import prepare_model_for_kbit_training
from peft import LoraConfig, LoraConfig, get_peft_model

class WhisperModelPrep:
    """
    Facilitates the preparation of datasets for fine-tuning with the Whisper model,
    encompassing feature extraction, tokenization, and dataset processing.
    
    Attributes:
        dataset (DatasetDict): Dataset to be prepared for the Whisper model.
        model_id (str): Identifier for the Whisper model to be utilized.
        language_code (str): ISO code representing the dataset's language.
        processing_task (str): Specific task for the Whisper model to execute.
    """
    
    def __init__(self, dataset: DatasetDict, model_id: str, language_abbr: str, processing_task: str, use_peft: bool):
        """
        Sets up the dataset and configuration for processing with the Whisper model.
        
        Parameters:
            dataset (DatasetDict): The dataset to be prepared.
            model_id (str, optional): Whisper model identifier.
            language_code (str, optional): Dataset language ISO code.
            processing_task (str, optional): Task for the Whisper model.
        """
        self.dataset = dataset
        self.model_id = model_id
        self.language_abbr = language_abbr
        self.processing_task = processing_task
        self.use_peft = use_peft
        
    
    def initialize_feature_extractor(self) -> WhisperFeatureExtractor:
        """
        Creates and retrieves a feature extractor based on the Whisper model.
        
        Returns:
            WhisperFeatureExtractor: Configured feature extractor for the model.
        """
        return WhisperFeatureExtractor.from_pretrained(self.model_id, cache_dir = f'./{self.language_abbr}/feature-extractor')
    
    def initialize_tokenizer(self) -> WhisperTokenizer:
        """
        Creates and retrieves a tokenizer for the Whisper model.
        
        Returns:
            WhisperTokenizer: Configured tokenizer for the model.
        """
        return WhisperTokenizer.from_pretrained(self.model_id, cache_dir=f'./{self.language_abbr}/tokenizer')

    def initialize_processor(self) -> WhisperProcessor:
        """
        Combines the feature extractor and tokenizer to create a processor for the Whisper model,
        facilitating seamless data preparation.
        
        Returns:
            WhisperProcessor: Unified processor for the model.
        """
        return WhisperProcessor.from_pretrained(self.model_id, self.language_abbr, self.processing_task)
     
    def initialize_model(self) -> WhisperForConditionalGeneration:
        """
        Initializes and retrieves the Whisper model configured for conditional generation.

        This method sets up the Whisper model with specific configurations, ensuring it is
        ready for use in tasks such as transcription or translation, depending on the
        processing task specified during the class initialization.

        Returns:
            WhisperForConditionalGeneration: The configured Whisper model ready for conditional generation tasks.
        """
        
        if self.use_peft:
            model = WhisperForConditionalGeneration.from_pretrained(self.model_id, cache_dir=f'./{self.language_abbr}/model', load_in_8bit=True, device_map="auto")
            model.config.forced_decoder_ids = None
            model.config.suppress_tokens = []  
            model.generation_config.language = "en"  
            model.generation_config.task = "transcribe"
            model = prepare_model_for_kbit_training(model)
            config = LoraConfig(r=32, lora_alpha=64, target_modules=["q_proj", "v_proj"], lora_dropout=0.05, bias="none")
            model = get_peft_model(model, config)
            model.print_trainable_parameters()
        else:
            print("PEFT optimization is not enabled.")
            model = WhisperForConditionalGeneration.from_pretrained(self.model_id, cache_dir=f'./{self.language_abbr}/model')
            model.config.forced_decoder_ids = None
            model.config.suppress_tokens = []  
            model.generation_config.language = "en"  
            model.generation_config.task = "transcribe"
        
        return model
    



from transformers import  WhisperFeatureExtractor, WhisperTokenizer, WhisperProcessor, WhisperForConditionalGeneration
from .load_data import Dataset
from .whisper_model_prep import WhisperModelPrep
from .audio_data_processor import AudioDataProcessor
from datasets import DatasetDict
from typing import Tuple

class DataPrep:
    """
    
    A class to encapsulate preparing the speech dataset for model training, including dataset loading, cleaning, and preprocessing tasks like feature extraction and tokenization. .

    """

    def __init__(self, huggingface_read_token: str, dataset_name: str, language_abbr: str, model_id: str, processing_task: str, use_peft: bool):
        """
        Initializes the Trainer with the necessary configuration and loads the evaluation metric.

        Parameters:
            huggingface_token (str): Hugging Face API token for authenticated access.
            dataset_name (str): Name of the dataset to be downloaded from Hugging Face.
            language_abbr (str): Language abbreviation for the dataset.
            model_id (str): Model ID for the model to be used in training.
            processing_task (str): The processing task to be performed (e.g., "transcribe").
        """
        self.huggingface_read_token = huggingface_read_token
        self.dataset_name = dataset_name
        self.language_abbr = language_abbr
        self.model_id = model_id
        self.processing_task = processing_task
        self.use_peft = use_peft
        self.model_prep = WhisperModelPrep(self.dataset_name, self.model_id, self.language_abbr, self.processing_task, self.use_peft)
        self.data_loader = Dataset(self.huggingface_read_token, self.dataset_name, self.language_abbr)

    def prepare_model(self) -> Tuple[WhisperFeatureExtractor, WhisperTokenizer, WhisperProcessor, WhisperForConditionalGeneration]:
        """Initializes and sets up the necessary components for model training, including the tokenizer, 
        feature extractor, feature processor, and the model itself. 
        
        Returns:
        tuple: A tuple containing four elements in the order of tokenizer, feature extractor, 
               feature processor, and the model. Each element is an initialized instance of its 
               respective class, ready for use in the training pipeline.
        """
        
        self.tokenizer = self.model_prep.initialize_tokenizer()
        self.feature_extractor = self.model_prep.initialize_feature_extractor()
        self.feature_processor = self.model_prep.initialize_processor()
        self.model = self.model_prep.initialize_model()
        return self.tokenizer, self.feature_extractor, self.feature_processor, self.model
        

    def load_dataset(self, feature_extractor, tokenizer)-> DatasetDict:
        """
        Retrieves and preprocesses the specified dataset for model training and evaluation.

        Parameters:
        feature_extractor (PreTrainedFeatureExtractor): The feature extractor instance to be used for 
                                                        audio data preprocessing.
        tokenizer (PreTrainedTokenizer): The tokenizer instance for processing textual data associated 
                                         with the audio samples.

        Returns:
            DatasetDict: A dictionary containing the preprocessed 'train' and 'test' splits of the dataset, 
                        ready for use in model training and evaluation. Each split has been cleaned and 
                        processed to include only the necessary features for model input.
        """
        self.dataset = self.data_loader.load_dataset()
        print(f"Training dataset size: {len(self.dataset['train'])}")
        print(f"Test dataset size: {len(self.dataset['test'])}")
        processor = AudioDataProcessor(self.dataset, feature_extractor, tokenizer)
        cleaned_dataset = processor.clean_dataset()
        cleaned_dataset["train"] = cleaned_dataset["train"].map(lambda example: processor.resampled_dataset(example))
        cleaned_dataset["test"] = cleaned_dataset["test"].map(lambda example: processor.resampled_dataset(example))
        return cleaned_dataset

        









from datasets import DatasetDict
import librosa
from transformers import PreTrainedTokenizer

class AudioDataProcessor:
    """
    Processes audio datasets for use in whisper models, including cleaning and resampling.

    """
    
    def __init__(self, dataset: DatasetDict, feature_extractor, tokenizer: PreTrainedTokenizer):
        """
        Initializes the DatasetProcessor with the dataset, feature extractor, and tokenizer.

        Parameters:
            dataset (DatasetDict): The dataset to process.
            feature_extractor (PreTrainedFeatureExtractor): The feature extractor for audio data.
            tokenizer (PreTrainedTokenizer): The tokenizer for text data.
        """
        self.dataset = dataset
        self.feature_extractor = feature_extractor
        self.tokenizer = tokenizer

    def clean_dataset(self) -> DatasetDict:
        """
        Cleans the dataset by removing unnecessary columns to streamline processing.

        Returns:
            DatasetDict: The cleaned dataset.
        """
        columns_to_remove = ["accent", "age", "client_id", "down_votes", "gender", "locale", "path", "segment", "up_votes"]
        self.dataset = self.dataset.remove_columns(columns_to_remove)
        return self.dataset

    def resampled_dataset(self, sample) -> DatasetDict:
        """
        Resamples the audio data to the required sampling rate and extracts features using the feature extractor.

        Parameters:
            example (dict): A single sample from the dataset, containing 'audio' and 'sentence' keys.

        Returns:
            dict: The updated sample resampled to 16000kHz.
        """
        resampled_audio = librosa.resample(sample["audio"]["array"], orig_sr=sample["audio"]["sampling_rate"], target_sr=16000)

        sample["audio"]["array"] = resampled_audio
        sample["audio"]["sampling_rate"] = 16000

        audio_features = self.feature_extractor(resampled_audio, sampling_rate=16000).input_features[0]

        tokenized_sentence = self.tokenizer(sample["sentence"]).input_ids

        sample["input_features"] = audio_features
        sample["labels"] = tokenized_sentence

        return sample

        







from fastapi import FastAPI, File, UploadFile, HTTPException
from pydantic import BaseModel
from typing import Optional
import pytube as pt
from transformers import pipeline, WhisperTokenizer
import torch
import tempfile

app = FastAPI()

model_name = " "
tokenizer = " "
huggingface_read_token = " "
language_abbr = " "

tokenizer = WhisperTokenizer.from_pretrained(tokenizer)
device = 0 if torch.cuda.is_available() else "cpu"
pipe = pipeline(
    task = "automatic-speech-recognition",
    model = model_name,
    token = huggingface_read_token,
    tokenizer=tokenizer,
    device=device
)
pipe.model.config.forced_decoder_ids = pipe.tokenizer.get_decoder_prompt_ids(language = language_abbr, task="transcribe")


class AudioTranscriptionRequest(BaseModel):
    file: UploadFile

@app.post('/transcribe')
async def transcribe(file: UploadFile):
    if file is None:
        raise HTTPException(status_code=400, detail="No file provided")

    try:
        with tempfile.NamedTemporaryFile(suffix='.tmp', delete=False) as tmp_file:
            tmp_file.write(await file.read())
            tmp_file_path = tmp_file.name

        if file.filename.endswith('.mp3'):
            text = pipe(tmp_file_path)["text"]
        elif file.filename.endswith('.wav'):
            pass
        else:
            raise HTTPException(status_code=400, detail="Unsupported file format")

        return {'text': text}
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Error processing file: {str(e)}")
    



import unittest
from unittest.mock import patch, MagicMock
import sys
sys.path.append("load_data")
from src.load_data import  Dataset

# from load_data import Dataset

class TestDataset(unittest.TestCase):
    def setUp(self):
        self.huggingface_token = "your_token"
        self.dataset_name = "your_dataset_name"
        self.language_abbr = "your_language_abbr"
        self.dataset_manager = Dataset(self.huggingface_token, self.dataset_name, self.language_abbr)
        
    @patch("os.path.exists")
    @patch("datasets.load_dataset")
    def test_load_dataset_from_cache(self, mock_load_dataset, mock_os_path_exists):
        mock_os_path_exists.return_value = True
        # Mock the returned dataset
        mock_dataset = MagicMock()
        mock_load_dataset.return_value = mock_dataset
        
        dataset_dict = self.dataset_manager.load_dataset()
        
        mock_load_dataset.assert_called_with(self.dataset_name, self.language_abbr, split="train", 
                                              token=self.huggingface_token, streaming=False, 
                                              trust_remote_code=True, cache_dir=f'./{self.language_abbr}')
        mock_load_dataset.assert_called_with(self.dataset_name, self.language_abbr, split="test", 
                                              token=self.huggingface_token, streaming=False, 
                                              trust_remote_code=True, cache_dir=f'./{self.language_abbr}')
        self.assertEqual(dataset_dict["train"], mock_dataset)
        self.assertEqual(dataset_dict["test"], mock_dataset)
        
    @patch("os.path.exists")
    @patch("datasets.load_dataset")
    def test_load_dataset_from_remote(self, mock_load_dataset, mock_os_path_exists):
        mock_os_path_exists.return_value = False
        # Mock the returned dataset
        mock_dataset = MagicMock()
        mock_load_dataset.return_value = mock_dataset
        
        dataset_dict = self.dataset_manager.load_dataset()
        
        mock_load_dataset.assert_called_with(self.dataset_name, self.language_abbr, split="train", 
                                              token=self.huggingface_token, streaming=False, 
                                              trust_remote_code=True, cache_dir=f'./{self.language_abbr}')
        mock_load_dataset.assert_called_with(self.dataset_name, self.language_abbr, split="test", 
                                              token=self.huggingface_token, streaming=False, 
                                              trust_remote_code=True, cache_dir=f'./{self.language_abbr}')
        self.assertEqual(dataset_dict["train"], mock_dataset)
        self.assertEqual(dataset_dict["test"], mock_dataset)
        
if __name__ == '__main__':
    unittest.main()



from transformers.integrations import WandbCallback
import holoviews as hv
hv.extension("bokeh", logo=False)
import panel as pn
import numpy as np
import wandb
import pandas as pd
import jiwer
import tempfile
import numbers
from pathlib import Path
from io import StringIO



class RecordAnalyzer:
    """
    Class to analyze and process records from a dataset.
    """
    def __init__(self):
        self.tokenizer = None

    def record_to_html(self, sample_record) -> StringIO:
        """
        Convert a sample record to HTML format.

        Args:
            sample_record (dict): A sample record containing audio and spectrogram data.

        Returns:
            StringIO: HTML representation of the sample record.
        """
        audio_array = np.array(sample_record["audio"]["array"])
        audio_sr = sample_record["audio"]["sampling_rate"]
        audio_data = sample_record['audio']
        audio_duration = len(audio_data) / 16000
        audio_spectrogram = np.array(sample_record["spectrogram"])

        bounds = (0, 0, audio_duration, audio_spectrogram.max())

        waveform_int = np.int16(audio_array * 32767)

        hv_audio = pn.pane.Audio(waveform_int, sample_rate=audio_sr, name='Audio', throttle=500)

        slider = pn.widgets.FloatSlider(end=audio_duration, visible=False, step=0.001)
        line_audio = hv.VLine(0).opts(color='black')
        line_spec = hv.VLine(0).opts(color='red')

        slider.jslink(hv_audio, value='time', bidirectional=True)
        slider.jslink(line_audio, value='glyph.location')
        slider.jslink(line_spec, value='glyph.location')

        time = np.linspace(0, audio_duration, num=len(audio_array))
        line_plot_hv = hv.Curve(
            (time, audio_array), ["Time (s)", "amplitude"]).opts(
            width=500, height=150, axiswise=True) * line_audio

        hv_spec_gram = hv.Image(
            audio_spectrogram, bounds=(bounds), kdims=["Time (s)", "Frequency (hz)"]).opts(
            width=500, height=150, labelled=[], axiswise=True, color_levels=512) * line_spec

        combined = pn.Row(hv_audio, hv_spec_gram, line_plot_hv, slider)
        audio_html = StringIO()
        combined.save(audio_html)
        return audio_html

    def dataset_to_records(self, dataset) -> pd.DataFrame:
        """
        Convert a dataset to a DataFrame of records.

        Args:
            dataset (list): List of records in the dataset.

        Returns:
            pd.DataFrame: DataFrame containing the processed records.
        """
        records = []
        for item in dataset:
            record = {}
            audio_data = item['audio']
            audio_duration = len(audio_data) / 16000
            record["audio_with_spec"] = wandb.Html(self.record_to_html(item))
            record["sentence"] = item["sentence"]
            record["length"] = audio_duration
            records.append(record)
        records = pd.DataFrame(records)
        return records

    def decode_predictions(self, predictions, tokenizer) -> list:
        """
        Decode model predictions into human-readable format.

        Args:
            predictions (object): Predictions generated by the model.

        Returns:
            list: Decoded predictions.
        """
        pred_ids = predictions.predictions
        pred_str = tokenizer.batch_decode(pred_ids, skip_special_tokens=True)
        return pred_str

    def compute_measures(self, predictions, labels) -> pd.DataFrame:
        """
        Compute evaluation measures for model predictions.

        Args:
            predictions (list): List of model predictions.
            labels (list): List of ground truth labels.

        Returns:
            pd.DataFrame: DataFrame containing computed evaluation measures.
        """
        measures = [jiwer.compute_measures(ls, ps) for ps, ls in zip(predictions, labels)]
        measures_df = pd.DataFrame(measures)[["wer", "hits", "substitutions", "deletions", "insertions"]]
        return measures_df




class WandbProgressResultsCallback(WandbCallback):
    """
    Callback class for logging training progress to Weights & Biases.
    """
    def __init__(self, trainer, sample_dataset, tokenizer): 
        """
        Initialize the WandbProgressResultsCallback instance.

        Args:
            trainer: The instance of the Trainer class.
            sample_dataset: The dataset used for logging sample predictions.
            tokenizer: The tokenizer instance for decoding predictions.
        """
        super().__init__()
        self.trainer = trainer
        self.sample_dataset = sample_dataset
        self.records_analyzer = RecordAnalyzer()
        self.records_df = self.records_analyzer.dataset_to_records(sample_dataset)
        self.tokenizer = tokenizer
        
    def on_log(self, args, state, control, model=None, logs=None, **kwargs):
        """
        Log training progress to Weights & Biases.

        Args:
            args: Arguments passed to the callback.
            state: The current state of the training.
            control: Control parameters for the callback.
            model: The trained model.
            logs: Additional logs.
            **kwargs: Additional keyword arguments.

        Returns:
            The table containing sample predictions.
        """
        super().on_log(args, state, control, model, logs)
        predictions = self.trainer.predict(self.sample_dataset)
        predictions = self.records_analyzer.decode_predictions(predictions, self.tokenizer)
        measures_df = self.records_analyzer.compute_measures(predictions, self.records_df["sentence"].tolist())
        records_df = pd.concat([self.records_df, measures_df], axis=1)
        records_df["prediction"] = predictions
        records_df["step"] = state.global_step
        records_table = self._wandb.Table(dataframe=records_df)
        self._wandb.log({"sample_predictions": records_table})
        
    def on_save(self, args, state, control, model=None, tokenizer=None, **kwargs):
        """
        Save the trained model as an artifact in Weights & Biases.

        Args:
            args: Arguments passed to the callback.
            state: The current state of the training.
            control: Control parameters for the callback.
            model: The trained model.
            tokenizer: The tokenizer instance.
            **kwargs: Additional keyword arguments.
        """
        if self._wandb is None:
            return
        if self._log_model and self._initialized and state.is_world_process_zero:
            with tempfile.TemporaryDirectory() as temp_dir:
                self.trainer.save_model(temp_dir)
                metadata = (
                    {
                        k: v
                        for k, v in dict(self._wandb.summary).items()
                        if isinstance(v, numbers.Number) and not k.startswith("_")
                    }
                    if not args.load_best_model_at_end
                    else {
                        f"eval/{args.metric_for_best_model}": state.best_metric,
                        "train/total_floss": state.total_flos,
                    }
                )
                artifact = self._wandb.Artifact(
                    name=f"model-{self._wandb.run.id}",
                    type="model", metadata=metadata)
                for f in Path(temp_dir).glob("*"):
                    if f.is_file():
                        with artifact.new_file(f.name, mode="wb") as fa:
                            fa.write(f.read_bytes())
                self._wandb.run.log_artifact(artifact)


import os
from datasets import load_dataset, DatasetDict

class Dataset:
    """
    Manages loading of the datasets from Hugging Face's Common Voice repository.
    
    Attributes:
        huggingface_token (str): Hugging Face API token for read authenticated access.
        dataset_name (str): Name of the dataset to be downloaded from Hugging Face.
        language_abbr (str): Abbreviation of the language for the dataset.
    """
    
    def __init__(self, huggingface_token: str, dataset_name: str, language_abbr: str):
        """
        Initializes the DatasetManager with necessary details for dataset operations.
        
        Parameters:
            huggingface_token (str): Hugging Face API token.
            dataset_name (str): Name of the dataset.
            language_abbr (str): Language abbreviation for the dataset.
        """
        self.huggingface_token = huggingface_token
        self.dataset_name = dataset_name
        self.language_abbr = language_abbr
    
    def load_dataset(self) -> DatasetDict:
        """
        Checks if the dataset is available in the cache directory. If so, loads it from there.
        Otherwise, downloads the specified dataset from Hugging Face, including 'train' and 'test' splits.

        Returns:
            DatasetDict: The downloaded dataset.
        """
        cache_directory = f'./{self.language_abbr}'
        if os.path.exists(cache_directory):
            print(f"Loading dataset for {self.language_abbr} from cache.")
        else:
            print(f"Cache directory not found for {self.language_abbr}. Downloading dataset...")

        common_voice = DatasetDict()
        common_voice["train"] = load_dataset(self.dataset_name, self.language_abbr, split="train", 
                                             token=self.huggingface_token, streaming=False, 
                                             trust_remote_code=True, cache_dir=cache_directory)
        common_voice["test"] = load_dataset(self.dataset_name, self.language_abbr, split="test", 
                                            token=self.huggingface_token, streaming=False, 
                                            trust_remote_code=True, cache_dir=cache_directory)

        return common_voice
    




import argparse
from .data_prep import DataPrep
from .model_trainer import Trainer

def parse_args():
    parser = argparse.ArgumentParser(description="Run the training orchestrator with specified parameters.")
    parser.add_argument("--huggingface_read_token", type=str, required=True, help="Hugging Face API token for read authenticated access.")
    parser.add_argument("--huggingface_write_token", type=str, required=True, help="Hugging Face API token for write authenticated access.")
    parser.add_argument("--dataset_name", type=str, default="mozilla-foundation/common_voice_16_1", help="Name of the dataset to be downloaded from Hugging Face.")
    parser.add_argument("--language_abbr", type=str, default="sw", help="Abbreviation of the language for the dataset.")
    parser.add_argument("--model_id", type=str, default="openai/whisper-small", help="Model ID for the model to be used in training.")
    parser.add_argument("--processing_task", type=str, default="transcribe", help="The processing task to be performed.")
    parser.add_argument("--wandb_api_key", type=str, help="The wandb.ai api key for monitoring training runs, signup and generate an api key.")
    parser.add_argument("--use_peft", type = bool, help = "True to train your model using PEFT method, False for full finetuning")

    return parser.parse_args()

if __name__ == "__main__":
    args = parse_args()
    process = DataPrep(
        huggingface_read_token=args.huggingface_read_token,
        dataset_name=args.dataset_name,
        language_abbr=args.language_abbr,
        model_id=args.model_id,
        processing_task=args.processing_task,
        use_peft = args.use_peft
    )
    tokenizer, feature_extractor, feature_processor, model = process.prepare_model()
    processed_dataset = process.load_dataset(feature_extractor, tokenizer)
    trainer = Trainer(
            huggingface_write_token=args.huggingface_write_token,
            model_id=args.model_id,
            dataset=processed_dataset,
            model=model,
            feature_processor=feature_processor,
            feature_extractor=feature_extractor,
            tokenizer=tokenizer,
            language_abbr=args.language_abbr,
            wandb_api_key=args.wandb_api_key,
            use_peft = args.use_peft
            )
    trainer.train()
    
    
    
